# Core
from .core import *

# MachineLearning
from .MachineLearning import *

__version__ = '0.1.3b'
__author__ = 'Zeesain Tsui'
__help__ = """
===============================================
Hello, this is a brief help sheet for oneline.
===============================================

# What is oneline?

oneline is a personal code set for learning and researching, and it provides advanced encapsulation of some libraries,
designing to provide a simpler mode of operation and help people to simplify their codes concisely.

# For example

from oneLine import OneData
data = OneData('test.csv')

Directory or array etc will be converted to OneData, and you can also use OneData as pd.DataFrame.

Have fun!
"""
